# -*- coding: utf-8 -*-
import logging
from datetime import datetime
from operator import itemgetter
from typing import Dict, Any, Optional, List, Tuple, Iterator, Union

import regex
from hdx.location.adminone import AdminOne

from hdx.utilities.dateparse import parse_date, get_datetime_from_timestamp
from hdx.utilities.dictandlist import dict_of_lists_add
from hdx.utilities.downloader import Download
from hdx.utilities.text import number_format, get_fraction_str, get_numeric_if_possible

from hdx.scraper import get_rowval, add_population
from hdx.scraper.readers import read
from hdx.scraper.rowparser import RowParser

logger = logging.getLogger(__name__)

brackets = r'''
(?<rec> #capturing group rec
 \( #open parenthesis
 (?: #non-capturing group
  [^()]++ #anything but parenthesis one or more times without backtracking
  | #or
   (?&rec) #recursive substitute of group rec
 )*
 \) #close parenthesis
)'''


def _run_scraper(countryiso3s, adminone, level, name, datasetinfo, headers, iterator, population_lookup, retheaders=[list(), list()], retvalues=list(), sources=list()):
    # type: (List[str], AdminOne, str, str, Dict, List[str], Iterator[Union[List,Dict]], Dict[str,int], List, List, List) -> None
    """Run one mini scraper.

    Args:
        countryiso3s (List[str]): List of ISO3 country codes to process
        adminone (AdminOne): AdminOne object from HDX Python Country library that handles processing of admin level 1
        level (str): Can be global, national or subnational
        name (str): Name of mini scraper
        datasetinfo (Dict): Dictionary of information about dataset
        headers (List[str]): Row headers
        iterator (Iterator[Union[List,Dict]]): Rows
        population_lookup (Dict[str,int]): Dictionary from admin code to population
        retheaders (List): Headers to output. Defaults to [list(), list()].
        retvalues (List): Values to output. Defaults to list().
        sources (List): Sources to output. Defaults to list().

    Returns:
        Tuple[Optional[str], Optional[List[bool]]]: (admin name, should process subset list) or (None, None)
    """
    subsets = datasetinfo.get('subsets')
    if not subsets:
        subsets = [{'filter': datasetinfo.get('filter'), 'input_cols': datasetinfo.get('input_cols', list()),
                          'input_transforms': datasetinfo.get('input_transforms', dict()), 'process_cols':  datasetinfo.get('process_cols', list()),
                          'input_keep': datasetinfo.get('input_keep', list()), 'input_append': datasetinfo.get('input_append', list()),
                          'sum_cols': datasetinfo.get('sum_cols'), 'input_ignore_vals': datasetinfo.get('input_ignore_vals', list()),
                          'output_cols': datasetinfo.get('output_cols', list()), 'output_hxltags': datasetinfo.get('output_hxltags', list())}]
    use_hxl = datasetinfo.get('use_hxl', False)
    if use_hxl:
        hxlrow = next(iterator)
        while not hxlrow:
            hxlrow = next(iterator)
        exclude_tags = datasetinfo.get('exclude_tags', list())
        adm_cols = list()
        input_cols = list()
        columns = list()
        for header in headers:
            hxltag = hxlrow[header]
            if not hxltag or hxltag in exclude_tags:
                continue
            if '#country' in hxltag:
                if 'code' in hxltag:
                    if len(adm_cols) == 0:
                        adm_cols.append(hxltag)
                    else:
                        adm_cols[0] = hxltag
                continue
            if '#adm1' in hxltag:
                if 'code' in hxltag:
                    if len(adm_cols) == 0:
                        adm_cols.append(None)
                    if len(adm_cols) == 1:
                        adm_cols.append(hxltag)
                continue
            if hxltag == datasetinfo.get('date_col') and datasetinfo.get('include_date', False) is False:
                continue
            input_cols.append(hxltag)
            columns.append(header)
        datasetinfo['adm_cols'] = adm_cols
        for subset in subsets:
            orig_input_cols = subset.get('input_cols', list())
            if not orig_input_cols:
                orig_input_cols.extend(input_cols)
            subset['input_cols'] = orig_input_cols
            orig_columns = subset.get('output_cols', list())
            if not orig_columns:
                orig_columns.extend(columns)
            subset['output_cols'] = orig_columns
            orig_hxltags = subset.get('output_hxltags', list())
            if not orig_hxltags:
                orig_hxltags.extend(input_cols)
            subset['output_hxltags'] = orig_hxltags
    else:
        hxlrow = None

    rowparser = RowParser(countryiso3s, adminone, level, datasetinfo, headers, subsets)
    valuedicts = dict()
    for subset in subsets:
        for _ in subset['input_cols']:
            dict_of_lists_add(valuedicts, subset['filter'], dict())

    def add_row(row):
        adm, should_process_subset = rowparser.parse(row, name)
        if not adm:
            return
        for i, subset in enumerate(subsets):
            if not should_process_subset[i]:
                continue
            filter = subset['filter']
            input_ignore_vals = subset.get('input_ignore_vals', list())
            input_transforms = subset.get('input_transforms', dict())
            sum_cols = subset.get('sum_cols')
            process_cols = subset.get('process_cols')
            input_append = subset.get('input_append', list())
            input_keep = subset.get('input_keep', list())
            for i, valcol in enumerate(subset['input_cols']):
                valuedict = valuedicts[filter][i]
                val = get_rowval(row, valcol)
                input_transform = input_transforms.get(valcol)
                if input_transform and val not in input_ignore_vals:
                    val = eval(input_transform.replace(valcol, 'val'))
                if sum_cols or process_cols:
                    dict_of_lists_add(valuedict, adm, val)
                else:
                    curval = valuedict.get(adm)
                    if valcol in input_append:
                        if curval:
                            val = curval + val
                    elif valcol in input_keep:
                        if curval:
                            val = curval
                    valuedict[adm] = val

    stop_row = datasetinfo.get('stop_row')
    for row in iterator:
        if not isinstance(row, dict):
            row = row.value
        if hxlrow:
            newrow = dict()
            for header in row:
                newrow[hxlrow[header]] = row[header]
            row = newrow
        if stop_row:
            if all(row[key] == value for key, value in stop_row.items()):
                break
        for newrow in rowparser.flatten(row):
            add_row(newrow)

    date = datasetinfo.get('date')
    use_date_from_date_col = datasetinfo.get('use_date_from_date_col', False)
    if date and not use_date_from_date_col:
        date = parse_date(date)
    else:
        date = rowparser.get_maxdate()
        if date == 0:
            raise ValueError('No date given in datasetinfo or as a column!')
        if rowparser.datetype == 'date':
            if not isinstance(date, datetime):
                date = parse_date(date)
        elif rowparser.datetype == 'int':
            date = get_datetime_from_timestamp(date)
        else:
            raise ValueError('No date type specified!')
    date = date.strftime('%Y-%m-%d')

    for subset in subsets:
        output_cols = subset['output_cols']
        retheaders[0].extend(output_cols)
        output_hxltags = subset['output_hxltags']
        retheaders[1].extend(output_hxltags)
        valdicts = valuedicts[subset['filter']]
        process_cols = subset.get('process_cols')
        input_keep = subset.get('input_keep', list())
        sum_cols = subset.get('sum_cols')
        input_ignore_vals = subset.get('input_ignore_vals', list())
        valcols = subset['input_cols']
        # Indices of list sorted by length
        sorted_len_indices = sorted(range(len(valcols)), key=lambda k: len(valcols[k]), reverse=True)
        if process_cols:
            newvaldicts = [dict() for _ in process_cols]

            def text_replacement(string, adm):
                string = string.replace('#population', '#pzbgvjh')
                hasvalues = False
                for j in sorted_len_indices:
                    valcol = valcols[j]
                    if valcol not in string:
                        continue
                    if valcol in input_keep:
                        input_keep_index = 0
                    else:
                        input_keep_index = -1
                    val = valdicts[j][adm][input_keep_index]
                    if val is None or val == '' or val in input_ignore_vals:
                        val = 0
                    else:
                        hasvalues = True
                    string = string.replace(valcol, str(val))
                string = string.replace('#pzbgvjh', '#population')
                return string, hasvalues

            for i, process_col in enumerate(process_cols):
                valdict0 = valdicts[0]
                for adm in valdict0:
                    hasvalues = True
                    matches = regex.search(brackets, process_col, flags=regex.VERBOSE)
                    if matches:
                        for bracketed_str in matches.captures('rec'):
                            if any(bracketed_str in x for x in valcols):
                                continue
                            _, hasvalues_t = text_replacement(bracketed_str, adm)
                            if not hasvalues_t:
                                hasvalues = False
                                break
                    if hasvalues:
                        formula, hasvalues_t = text_replacement(process_col, adm)
                        if hasvalues_t:
                            formula = formula.replace('#population', 'population_lookup[adm]')
                            newvaldicts[i][adm] = eval(formula)
                        else:
                            newvaldicts[i][adm] = ''
                    else:
                        newvaldicts[i][adm] = ''
            retvalues.extend(newvaldicts)
        elif sum_cols:
            for sum_col in sum_cols:
                formula = sum_col['formula']
                mustbepopulated = sum_col.get('mustbepopulated', False)
                newvaldicts = [dict() for _ in valdicts]
                valdict0 = valdicts[0]
                for adm in valdict0:
                    for i, val in enumerate(valdict0[adm]):
                        if not val or val in input_ignore_vals:
                            exists = False
                        else:
                            exists = True
                            for valdict in valdicts[1:]:
                                val = valdict[adm][i]
                                if val is None or val == '' or val in input_ignore_vals:
                                    exists = False
                                    break
                        if mustbepopulated and not exists:
                            continue
                        for j, valdict in enumerate(valdicts):
                            newvaldicts[j][adm] = eval(f'newvaldicts[j].get(adm, 0.0) + {str(valdict[adm][i])}')
                formula = formula.replace('#population', '#pzbgvjh')
                for i in sorted_len_indices:
                    formula = formula.replace(valcols[i], 'newvaldicts[%d][adm]' % i)
                formula = formula.replace('#pzbgvjh', 'population_lookup[adm]')
                newvaldict = dict()
                for adm in valdicts[0].keys():
                    try:
                        val = eval(formula)
                    except (ValueError, TypeError, KeyError):
                        val = ''
                    newvaldict[adm] = val
                retvalues.append(newvaldict)
        else:
            retvalues.extend(valdicts)
        source = datasetinfo['source']
        if isinstance(source, str):
            source = {'default_source': source}
        source_url = datasetinfo['source_url']
        if isinstance(source_url, str):
            source_url = {'default_url': source_url}
        sources.extend([(hxltag, date, source.get(hxltag, source['default_source']), source_url.get(hxltag, source_url['default_url'])) for hxltag in output_hxltags])
    logger.info('Processed %s' % name)


def run_scrapers(configuration, countryiso3s, adminone, level, maindownloader, basic_auths=dict(), today=None, scrapers=None, population_lookup=None, **kwargs):
    # type: (Dict, List[str], AdminOne, str, Download, Dict[str,str], Optional[datetime], Optional[List[str]], Dict[str,int], Any) -> Tuple[List,List,List]
    """Runs all mini scrapers given in configuration and returns headers, values and sources.

    Args:
        configuration (Dict): Configuration for mini scrapers
        countryiso3s (List[str]): List of ISO3 country codes to process
        adminone (AdminOne): AdminOne object from HDX Python Country library that handles processing of admin level 1
        level (str): Can be global, national or subnational
        maindownloader (Download): Download object for downloading files
        basic_auths (Dict[str,str]): Dictionary of basic authentication information
        today (Optional[datetime]): Value to use for today. Defaults to None (datetime.now()).
        scrapers (Optional[List[str]])): List of mini scraper names to process
        population_lookup (Dict[str,int]): Dictionary from admin code to population
        **kwargs: Variables to use when evaluating template arguments in urls

    Returns:
        Tuple[List,List,List]: Output headers, values and sources
    """
    datasets = configuration['scraper_%s' % level]
    retheaders = [list(), list()]
    retvalues = list()
    sources = list()
    now = datetime.now()
    for name in datasets:
        if scrapers:
            if not any(scraper in name for scraper in scrapers):
                continue
        else:
            if name == 'population':
                continue
        logger.info('Processing %s' % name)
        basic_auth = basic_auths.get(name)
        if basic_auth is None:
            downloader = maindownloader
        else:
            downloader = Download(basic_auth=basic_auth, rate_limit={'calls': 1, 'period': 0.1})
        datasetinfo = datasets[name]
        datasetinfo['name'] = name
        headers, iterator = read(downloader, datasetinfo, today=today, **kwargs)
        if 'source_url' not in datasetinfo:
            datasetinfo['source_url'] = datasetinfo['url']
        if 'date' not in datasetinfo or datasetinfo.get('force_date_today', False):
            today_str = kwargs.get('today_str')
            if not today_str:
                if today:
                    today_str = today.strftime('%Y-%m-%d')
                else:
                    today_str = now.strftime('%Y-%m-%d')
            datasetinfo['date'] = today_str
        sort = datasetinfo.get('sort')
        if sort:
            keys = sort['keys']
            reverse = sort.get('reverse', False)
            iterator = sorted(list(iterator), key=itemgetter(*keys), reverse=reverse)
        _run_scraper(countryiso3s, adminone, level, name, datasetinfo, headers, iterator, population_lookup, retheaders, retvalues, sources)
        if downloader != maindownloader:
            downloader.close()
        if population_lookup is not None:
            add_population(population_lookup, retheaders, retvalues)
    return retheaders, retvalues, sources



import filecmp
from os import getenv
from os.path import join

import pandas
import pytest
from hdx.utilities.dateparse import parse_date
from hdx.utilities.downloader import Download
from hdx.utilities.path import temp_dir

from hdx.scraper.exceloutput import ExcelOutput
from hdx.scraper.googlesheets import GoogleSheets
from hdx.scraper.jsonoutput import JsonOutput
from hdx.scraper.nooutput import NoOutput


class TestOutput:
    @pytest.fixture(scope='class')
    def hxltags(self):
        return {'Country Code': '#country+code', 'Country Name': '#country+name', 'Population': '#population'}

    def test_save(self, configuration, fixtures, hxltags):
        with temp_dir('TestScraperSave', delete_on_success=True, delete_on_failure=False) as tempdir:
            with Download(user_agent='test') as downloader:
                tabs = configuration['tabs']
                sheetname = list(tabs.values())[0]
                noout = NoOutput(tabs)
                excelout = ExcelOutput(join(tempdir, 'test_output.xlsx'), tabs, tabs)
                gsheet_auth = getenv('GSHEET_AUTH')
                if not gsheet_auth:
                    raise ValueError('No gsheet authorisation supplied!')
                googleout = GoogleSheets(configuration, gsheet_auth, None, tabs, tabs)
                jsonout = JsonOutput(configuration, tabs)
                output = [list(hxltags.keys()), list(hxltags.values()), ['AFG', 'Afghanistan', 38041754]]

                # won't do anything as wrong tab name
                excelout.update_tab('lala', output, hxltags=hxltags)
                googleout.update_tab('lala', output, hxltags=hxltags)
                jsonout.update_tab('lala', output, hxltags=hxltags)

                noout.update_tab('national', output, hxltags=hxltags)
                excelout.update_tab('national', output, hxltags=hxltags)
                googleout.update_tab('national', output, hxltags=hxltags)
                jsonout.update_tab('national', output, hxltags=hxltags)
                noout.add_additional_json(downloader, today=parse_date('2020-10-01'))
                jsonout.add_additional_json(downloader, today=parse_date('2020-10-01'))
                noout.save()
                excelout.save()
                filepaths = jsonout.save(tempdir, countries_to_save=['AFG'])
                excelsheet = excelout.workbook.get_sheet_by_name(sheetname)

                def get_list_from_cells(cells):
                    result = [list(), list(), list()]
                    for i, row in enumerate(excelsheet[cells]):
                        for column in row:
                            result[i].append(column.value)
                    return result

                assert get_list_from_cells('A1:C3') == output
                spreadsheet = googleout.gc.open_by_url(configuration['googlesheets']['test'])
                googletab = spreadsheet.worksheet_by_title(sheetname)
                result = googletab.get_values(start=(1, 1), end=(3, 3), returnas='matrix')
                result[2][2] = int(result[2][2])
                assert result == output
                assert filecmp.cmp(filepaths[0], join(fixtures, 'test_scraper_all.json'))
                assert filecmp.cmp(filepaths[1], join(fixtures, 'test_scraper_population.json'))
                assert filecmp.cmp(filepaths[2], join(fixtures, 'test_scraper_population.json'))
                assert filecmp.cmp(filepaths[3], join(fixtures, 'test_scraper_other.json'))

                jsonout.json = dict()
                df = pandas.DataFrame(output[2:], columns=output[0])
                noout.update_tab('national', df, hxltags=hxltags)
                excelout.update_tab('national', df, hxltags=hxltags)
                googleout.update_tab('national', df, hxltags=hxltags)
                jsonout.update_tab('national', df, hxltags=hxltags)
                jsonout.add_additional_json(downloader, today=parse_date('2020-10-01'))
                filepaths = jsonout.save(tempdir, countries_to_save=['AFG'])
                assert get_list_from_cells('A1:C3') == output
                result = googletab.get_values(start=(1, 1), end=(3, 3), returnas='matrix')
                result[2][2] = int(result[2][2])
                assert result == output
                assert filecmp.cmp(filepaths[0], join(fixtures, 'test_scraper_all.json'))
                assert filecmp.cmp(filepaths[1], join(fixtures, 'test_scraper_population.json'))
                assert filecmp.cmp(filepaths[2], join(fixtures, 'test_scraper_population.json'))
                assert filecmp.cmp(filepaths[3], join(fixtures, 'test_scraper_other.json'))

                df = pandas.DataFrame(output[1:], columns=output[0])
                googleout.update_tab('national', df, limit=2)
                result = googletab.get_values(start=(1, 1), end=(3, 3), returnas='matrix')
                result[2][2] = int(result[2][2])

    def test_jsonoutput(self, configuration, fixtures, hxltags):
        with temp_dir('TestScraperJson', delete_on_success=True, delete_on_failure=False) as tempdir:
            with Download(user_agent='test') as downloader:
                tabs = configuration['tabs']
                noout = NoOutput(tabs)
                jsonout = JsonOutput(configuration, tabs)
                rows = [{'Country Code': '#country+code', 'Country Name': '#country+name', 'Population': '#population'}, 
                        {'Country Code': 'AFG', 'Country Name': 'Afghanistan', 'Population': 38041754}]
                noout.add_data_rows_by_key('test', 'AFG', rows, hxltags)
                jsonout.add_data_rows_by_key('test', 'AFG', rows, hxltags)
                assert jsonout.json == {'test_data': {'AFG': [{'#country+code': '#country+code', '#country+name': '#country+name', '#population': '#population'},
                                                              {'#country+code': 'AFG', '#country+name': 'Afghanistan', '#population': 38041754}]}}
                df = pandas.DataFrame.from_records([rows[1]])
                jsonout.json = dict()
                noout.add_dataframe_rows('test', df, rows[0])
                jsonout.add_dataframe_rows('test', df, rows[0])
                assert jsonout.json == {'test_data': [{'#country+code': 'AFG', '#country+name': 'Afghanistan', '#population': 38041754}]}

                noout.add_data_row('test', rows[1])  # doesn't do anything



# -*- coding: utf-8 -*-
import copy
from datetime import datetime
from typing import List, Dict, Tuple, Iterator, Union, Generator, Optional

import hxl
from hdx.location.adminone import AdminOne
from hdx.location.country import Country
from hdx.utilities.dateparse import parse_date
from hdx.utilities.dictandlist import dict_of_lists_add

from hdx.scraper import match_template


class RowParser(object):
    """RowParser class for parsing each row.

    Args:
        countryiso3s (List[str]): List of ISO3 country codes to process
        adminone (AdminOne): AdminOne object from HDX Python Country library that handles processing of admin level 1
        level (str): Can be global, national or subnational
        datasetinfo (Dict): Dictionary of information about dataset
        headers (List[str]): Row headers
        subsets (List[Dict]): List of subset definitions
        maxdateonly (bool): Whether to only take the most recent date. Defaults to True.
    """

    def __init__(self, countryiso3s, adminone, level, datasetinfo, headers, subsets, maxdateonly=True):
        # type: (List[str], AdminOne, str, Dict, List[str], List[Dict], bool) -> None
        if isinstance(level, str):
            if level == 'global':
                level = None
            elif level == 'national':
                level = 0
            else:
                level = 1
        self.level = level
        self.datecol = datasetinfo.get('date_col')
        self.datetype = datasetinfo.get('date_type')
        if self.datetype:
            if self.datetype == 'date':
                date = parse_date('1900-01-01')
            else:
                date = 0
        else:
            date = 0
        self.maxdate = date
        date_condition = datasetinfo.get('date_condition')
        if date_condition is not None:
            for col in datasetinfo['input_cols']:
                date_condition = date_condition.replace(col, f"row['{col}']")
        self.date_condition = date_condition
        self.adminone = adminone
        self.admcols = datasetinfo.get('adm_cols', list())
        self.admexact = datasetinfo.get('adm_exact', False)
        self.subsets = subsets
        self.adms = [countryiso3s, self.adminone.pcodes]
        if self.level is None:
            self.maxdates = {i: date for i, _ in enumerate(subsets)}
        else:
            if self.level > len(self.admcols):
                raise ValueError('No admin columns specified for required level!')
            self.maxdates = {i: {adm: date for adm in self.adms[self.level]} for i, _ in enumerate(subsets)}

        self.maxdateonly = maxdateonly
        self.flatteninfo = datasetinfo.get('flatten')
        self.headers = headers
        self.filters = dict()
        self.read_external_filter(datasetinfo)

    def read_external_filter(self, datasetinfo):
        # type: (Dict) -> Tuple[List[str],Iterator[Union[List,Dict]]]
        """Read filter list from external url poitning to a HXLated file

        Args:
            datasetinfo (Dict): Dictionary of information about dataset

        Returns:
            None
        """
        external_filter = datasetinfo.get('external_filter')
        if not external_filter:
            return
        hxltags = external_filter['hxltags']
        data = hxl.data(external_filter['url'])
        for row in data:
            for hxltag in data.columns:
                if hxltag.display_tag in hxltags:
                    dict_of_lists_add(self.filters, hxltag.header, row.get('#country+code'))

    def flatten(self, row):
        # type: (Dict) -> Generator[Dict]
        """Flatten a wide spreadsheet format into a long one

        Args:
            row (Dict): Row to flatten

        Returns:
            Generator[Dict]: Flattened row(s)
        """
        if not self.flatteninfo:
            yield row
            return
        counters = [-1 for _ in self.flatteninfo]
        while True:
            newrow = copy.deepcopy(row)
            for i, flatten in enumerate(self.flatteninfo):
                colname = flatten['original']
                template_string, replace_string = match_template(colname)
                if not template_string:
                    raise ValueError('Column name for flattening lacks an incrementing number!')
                if counters[i] == -1:
                    counters[i] = int(replace_string)
                else:
                    replace_string = '%d' % counters[i]
                colname = colname.replace(template_string, replace_string)
                if colname not in row:
                    return
                newrow[flatten['new']] = row[colname]
                extracol = flatten.get('extracol')
                if extracol:
                    newrow[extracol] = colname
                counters[i] += 1
            yield newrow

    def get_maxdate(self):
        # type: () -> datetime
        """Get the most recent date of the rows so far

        Returns:
            datetime: Most recent date in processed rows
        """
        return self.maxdate

    def filtered(self, row):
        # type: (Dict) -> bool
        """Check if the row should be filtered out

        Args:
            row (Dict): Row to check for filters

        Returns:
            bool: Whether row is filtered out or not
        """
        for header in self.filters:
            if header not in row:
                continue
            if row[header] not in self.filters[header]:
                return True
        return False

    def parse(self, row, scrapername=None):
        # type: (Dict, str) -> Tuple[Optional[str], Optional[List[bool]]]
        """Parse row checking for valid admin information and if the row should be filtered out in each subset given
        its definition.

        Args:
            row (Dict): Row to parse

        Returns:
            Tuple[Optional[str], Optional[List[bool]]]: (admin name, should process subset list) or (None, None)
        """
        if self.filtered(row):
            return None, None

        adms = [None for _ in range(len(self.admcols))]

        def get_adm(admcol, i):
            template_string, match_string = match_template(admcol)
            if template_string:
                admcol = self.headers[int(match_string)]
            adm = row[admcol]
            if not adm:
                return False
            adm = adm.strip()
            adms[i] = adm
            if adm in self.adms[i]:
                return True
            exact = False
            if self.admexact:
                adms[i] = None
            else:
                if i == 0:
                    adms[i], exact = Country.get_iso3_country_code_fuzzy(adm)
                elif i == 1:
                    adms[i], exact = self.adminone.get_pcode(adms[0], adm, scrapername)
                if adms[i] not in self.adms[i]:
                    adms[i] = None
            return exact

        for i, admcol in enumerate(self.admcols):
            if admcol is None:
                continue
            if isinstance(admcol, str):
                admcol = [admcol]
            for admcl in admcol:
                exact = get_adm(admcl, i)
                if adms[i] and exact:
                    break
            if not adms[i]:
                return None, None

        should_process_subset = list()
        for subset in self.subsets:
            filter = subset['filter']
            process = True
            if filter:
                filters = filter.split('|')
                match = True
                for filterstr in filters:
                    filter = filterstr.split('=')
                    if row[filter[0]] != filter[1]:
                        match = False
                        break
                process = match
            should_process_subset.append(process)

        if self.datecol:
            if isinstance(self.datecol, list):
                dates = [str(row[x]) for x in self.datecol]
                date = ''.join(dates)
            else:
                date = row[self.datecol]
            if self.datetype == 'date':
                if not isinstance(date, datetime):
                    date = parse_date(date)
                date = date.replace(tzinfo=None)
            else:
                date = int(date)
            if self.date_condition:
                if eval(self.date_condition) is False:
                    return None, None
            for i, process in enumerate(should_process_subset):
                if not process:
                    continue
                if date > self.maxdate:
                    self.maxdate = date
                if self.level is None:
                    if self.maxdateonly:
                        if date < self.maxdates[i]:
                            should_process_subset[i] = False
                        else:
                            self.maxdates[i] = date
                    else:
                        self.maxdates[i] = date
                else:
                    if self.maxdateonly:
                        if date < self.maxdates[i][adms[self.level]]:
                            should_process_subset[i] = False
                        else:
                            self.maxdates[i][adms[self.level]] = date
                    else:
                        self.maxdates[i][adms[self.level]] = date
        if self.level is None:
            return 'global', should_process_subset
        return adms[self.level], should_process_subset



# -*- coding: utf-8 -*-
import json
import logging
from typing import Dict, List, Union, Optional

import pygsheets
from google.oauth2 import service_account
from pandas import DataFrame

logger = logging.getLogger()


class GoogleSheets:
    """GoogleSheets class enabling writing to Google spreadsheets.

    Args:
        configuration (Dict): Configuration for Google Sheets
        gsheet_auth (str): Authorisation for Google Sheets/Drive
        updatesheets (List[str]): List of spreadsheets to update (eg. prod, test)
        tabs (Dict[str, str]): Dictionary of mappings from internal name to spreadsheet tab name
        updatetabs (List[str]): Tabs to update
    """

    def __init__(self, configuration, gsheet_auth, updatesheets, tabs, updatetabs):
        # type: (Dict, str, List[str], Dict[str, str], List[str]) -> None
        info = json.loads(gsheet_auth)
        scopes = ['https://www.googleapis.com/auth/spreadsheets']
        credentials = service_account.Credentials.from_service_account_info(info, scopes=scopes)
        self.gc = pygsheets.authorize(custom_credentials=credentials)
        self.googlesheets = configuration['googlesheets']
        if updatesheets is None:
            updatesheets = self.googlesheets.keys()
            logger.info('Updating all spreadsheets')
        else:
            logger.info('Updating only these spreadsheets: %s' % updatesheets)
        self.updatesheets = updatesheets
        self.tabs = tabs
        self.updatetabs = updatetabs

    def update_tab(self, tabname, values, hxltags=None, limit=None):
        # type: (str, Union[List, DataFrame], Optional[Dict], Optional[int]) -> None
        """Update tab with values

        Args:
            tabname (str): Tab to update
            values (Union[List, DataFrame]): Either values in a list of dicts or a DataFrame
            hxltags (Optional[Dict]): HXL tag mapping. Defaults to None.
            limit (Optional[int]): Maximum number of rows to output

        Returns:
            None
        """
        if tabname not in self.updatetabs:
            return
        for sheet in self.googlesheets:
            if sheet not in self.updatesheets:
                continue
            url = self.googlesheets[sheet]
            spreadsheet = self.gc.open_by_url(url)

            tab = spreadsheet.worksheet_by_title(self.tabs[tabname])
            tab.clear(fields='*')
            if isinstance(values, list):
                tab.update_values('A1', values)
            else:
                headers = list(values.columns.values)
                tab.update_row(1, headers)
                if hxltags:
                    tab.update_row(2, [hxltags.get(header, '') for header in headers])
                    start = 3
                else:
                    start = 2
                if limit is not None:
                    values = values.head(limit)
                tab.set_dataframe(values, (start, 1), copy_head=False)



import pytest
from hdx.utilities.downloader import Download

from hdx.scraper.readers import read, get_url


class TestReaders:
    def test_read(self, configuration):
        url = get_url('http://{{var}}', var='hello')
        assert url == 'http://hello'
        with Download(user_agent='test') as downloader:
            datasetinfo = {'name': 'test', 'dataset': 'sahel-humanitarian-needs-overview', 'format': 'csv'}
            headers, iterator = read(downloader, datasetinfo, a='b')
            assert headers == ['Country', 'nutrition', 'Affected 2017', 'In Need 2017', 'Targeted 2017', '% targeted']
            assert next(iterator) == {'Country': '#country', 'nutrition': '#sector?', 'Affected 2017': '#affected', 'In Need 2017': '#inneed', 'Targeted 2017': '#targeted', '% targeted': '#targeted+percentage'}
            assert next(iterator) == {'Country': 'Burkina Faso', 'nutrition': 'MAM', 'Affected 2017': '433,412', 'In Need 2017': '433,412', 'Targeted 2017': '             _', '% targeted': '0'}
            assert datasetinfo == {'name': 'test', 'dataset': 'sahel-humanitarian-needs-overview', 'format': 'csv', 'headers': 1, 'date': '2016-09-01',
                                   'source': 'Multiple organisations', 'source_url': 'https://data.humdata.org/dataset/sahel-humanitarian-needs-overview',
                                   'url': 'https://data.humdata.org/dataset/47f6ef46-500f-421a-9fa2-fefd93facf95/resource/2527ac5b-66fe-46f0-8b9b-7086d2c4ddd3/download/hno-2017-sahel-nutrition.csv'}
            with pytest.raises(ValueError):
                datasetinfo = {'name': 'test', 'format': 'unknown'}
                read(downloader, datasetinfo)
            with pytest.raises(ValueError):
                datasetinfo = {'name': 'test', 'dataset': 'sahel-humanitarian-needs-overview', 'format': 'json'}
                read(downloader, datasetinfo)



# -*- coding: utf-8 -*-
import logging
from datetime import datetime
from os.path import join
from typing import Union, List, Dict, Optional, Any

from hdx.utilities.dictandlist import dict_of_lists_add
from hdx.utilities.downloader import Download
from hdx.utilities.saver import save_json
from hdx.utilities.text import get_numeric_if_possible
from pandas import DataFrame

from hdx.scraper import match_template
from hdx.scraper.readers import read

logger = logging.getLogger(__name__)


class JsonOutput:
    """JsonOutput class enabling writing to JSON files.

    Args:
        configuration (Dict): Configuration for Google Sheets
        updatetabs (List[str]): Tabs to update
    """

    def __init__(self, configuration, updatetabs):
        self.json_configuration = configuration['json']
        self.updatetabs = updatetabs
        self.json = dict()

    def add_data_row(self, key, row):
        # type: (str, Dict) -> None
        """Add row to JSON under a key

        Args:
            key (str): Key in JSON to update
            rows (List[Dict]): List of dictionaries

        Returns:
            None
        """
        dict_of_lists_add(self.json, '%s_data' % key, row)

    def add_dataframe_rows(self, key, df, hxltags=None):
        # type: (str, DataFrame, Optional[Dict]) -> None
        """Add rows from dataframe under a key

        Args:
            key (str): Key in JSON to update
            df (DataFrame): Dataframe containing rows
            hxltags (Optional[Dict]): HXL tag mapping. Defaults to None.

        Returns:
            None
        """
        if hxltags:
            df = df.rename(columns=hxltags)
        self.json['%s_data' % key] = df.to_dict(orient='records')

    def add_data_rows_by_key(self, key, countryiso, rows, hxltags=None):
        # type: (str, str, List[Dict], Optional[Dict]) -> None
        """Add rows under both a key and an ISO 3 country code subkey

        Args:
            key (str): Key in JSON to update
            countryiso (str): Country to use as subkey
            rows (List[Dict]): List of dictionaries
            hxltags (Optional[Dict]): HXL tag mapping. Defaults to None.

        Returns:
            None
        """
        fullname = '%s_data' % key
        jsondict = self.json.get(fullname, dict())
        jsondict[countryiso] = list()
        for row in rows:
            if hxltags:
                newrow = dict()
                for header, hxltag in hxltags.items():
                    newrow[hxltag] = row[header]
            else:
                newrow = row
            jsondict[countryiso].append(newrow)
        self.json[fullname] = jsondict

    def generate_json_from_list(self, key, rows):
        # type: (str, List[Dict]) -> None
        """Generate JSON from key and rows list

        Args:
            key (str): Key in JSON to update
            rows (List[Dict]): List of dictionaries

        Returns:
            None
        """
        hxltags = rows[1]
        for row in rows[2:]:
            newrow = dict()
            for i, hxltag in enumerate(hxltags):
                value = row[i]
                if value in [None, '']:
                    continue
                newrow[hxltag] = str(value)
            self.add_data_row(key, newrow)

    def generate_json_from_df(self, key, df, hxltags):
        # type: (str, DataFrame, Optional[Dict]) -> None
        """Generate JSON from key and dataframe

        Args:
            key (str): Key in JSON to update
            df (DataFrame): Dataframe containing rows
            hxltags (Optional[Dict]): HXL tag mapping. Defaults to None.

        Returns:
            None
        """
        for i, row in df.iterrows():
            newrow = dict()
            row = row.to_dict()
            for i, hxltag in enumerate(hxltags):
                value = row.get(hxltag)
                if value in [None, '']:
                    value = None
                newrow[hxltags.get(hxltag)] = value
            self.add_data_row(key, newrow)

    def update_tab(self, tabname, values, hxltags=None):
        # type: (str, Union[List, DataFrame], Optional[Dict]) -> None
        """Update tab with values

        Args:
            tabname (str): Tab to update
            values (Union[List, DataFrame]): Either values in a list of dicts or a DataFrame
            hxltags (Optional[Dict]): HXL tag mapping. Defaults to None.

        Returns:
            None
        """
        if tabname not in self.updatetabs:
            return
        if isinstance(values, list):
            self.generate_json_from_list(tabname, values)
        else:
            # isinstance(values, DataFrame)
            self.generate_json_from_df(tabname, values, hxltags)

    def add_additional_json(self, downloader, today=None):
        # type: (Download, Optional[datetime]) -> None
        """Download JSON files and add them under keys defined in the configuration

        Args:
            downloader (Download): Download object for downloading JSON
            today (Optional[datetime]): Value to use for today. Defaults to None (datetime.now()).

        Returns:
            None
        """
        for datasetinfo in self.json_configuration.get('additional_json', list()):
            headers, iterator = read(downloader, datasetinfo, today=today)
            hxlrow = next(iterator)
            if not isinstance(hxlrow, dict):
                hxlrow = hxlrow.value
            name = datasetinfo['name']
            for row in iterator:
                newrow = dict()
                if not isinstance(row, dict):
                    row = row.value
                for key in row:
                    hxltag = hxlrow[key]
                    if hxltag != '':
                        newrow[hxlrow[key]] = row[key]
                self.add_data_row(name, newrow)

    def save(self, folder=None, **kwargs):
        # type: (str, Any) -> List[str]
        """Save JSON file and any addition subsets of that JSON defined in the additional configuration

        Args:
            folder (Optional[str]): Key in JSON to update
            **kwargs: Variables to use when evaluating template arguments

        Returns:
            List[str]: List of file paths
        """
        filepaths = list()
        filepath = self.json_configuration['filepath']
        if folder:
            filepath = join(folder, filepath)
        logger.info('Writing JSON to %s' % filepath)
        save_json(self.json, filepath)
        filepaths.append(filepath)
        for kwarg in kwargs:
            exec('%s=%s' % (kwarg, kwargs[kwarg]))
        additional = self.json_configuration.get('additional', list())
        for filedetails in additional:
            json = dict()
            remove = filedetails.get('remove')
            if remove is None:
                tabs = filedetails['tabs']
            else:
                tabs = list()
                for key in self.json.keys():
                    tab = key.replace('_data', '')
                    if tab not in remove:
                        tabs.append({'tab': tab})
            for tabdetails in tabs:
                key = f'{tabdetails["tab"]}_data'
                newjson = self.json.get(key)
                filters = tabdetails.get('filters', dict())
                hxltags = tabdetails.get('hxltags')
                if (filters or hxltags or remove) and isinstance(newjson, list):
                    rows = list()
                    for row in newjson:
                        ignore_row = False
                        for filter, allowed_values in filters.items():
                            value = row.get(filter)
                            if value:
                                if isinstance(allowed_values, str):
                                    template_string, match_string = match_template(allowed_values)
                                    if template_string:
                                        allowed_values = eval(allowed_values.replace(template_string, match_string))
                                if isinstance(allowed_values, list):
                                    if value not in allowed_values:
                                        ignore_row = True
                                        break
                                elif value != allowed_values:
                                    ignore_row = True
                                    break
                        if ignore_row:
                            continue
                        if hxltags is None:
                            newrow = row
                        else:
                            newrow = dict()
                            for hxltag in hxltags:
                                if hxltag in row:
                                    newrow[hxltag] = row[hxltag]
                        rows.append(newrow)
                    newjson = rows
                newkey = tabdetails.get('key', key)
                json[newkey] = newjson
            if not json:
                continue
            filedetailspath = filedetails['filepath']
            if folder:
                filedetailspath = join(folder, filedetailspath)
            logger.info('Writing JSON to %s' % filedetailspath)
            save_json(json, filedetailspath)
            filepaths.append(filedetailspath)
        return filepaths



# -*- coding: utf-8 -*-
from datetime import datetime
from typing import List, Union, Optional, Dict, Any

from hdx.utilities.downloader import Download
from pandas import DataFrame


class NoOutput:
    """NoOutput class for testing that mimics the format of other output classes but does nothing.

    Args:
        updatetabs (List[str]): Tabs to update
    """

    def __init__(self, updatetabs):
        # type: (List[str]) -> None
        self.updatetabs = updatetabs

    def update_tab(self, tabname, values, hxltags=None, limit=None):
        # type: (str, Union[List, DataFrame], Optional[Dict], Optional[int]) -> None
        """Update tab with values

        Args:
            tabname (str): Tab to update
            values (Union[List, DataFrame]): Either values in a list of dicts or a DataFrame
            hxltags (Optional[Dict]): HXL tag mapping. Defaults to None.
            limit (Optional[int]): Maximum number of rows to output

        Returns:
            None
        """
        return

    def add_data_row(self, key, row):
        # type: (str, Dict) -> None
        """Add row

        Args:
            key (str): Key to update
            rows (List[Dict]): List of dictionaries

        Returns:
            None
        """
        return

    def add_dataframe_rows(self, key, df, hxltags=None):
        # type: (str, DataFrame, Optional[Dict]) -> None
        """Add rows from dataframe under a key

        Args:
            key (str): Key in JSON to update
            df (DataFrame): Dataframe containing rows
            hxltags (Optional[Dict]): HXL tag mapping. Defaults to None.

        Returns:
            None
        """
        return

    def add_data_rows_by_key(self, name, countryiso, rows, hxltags=None):
        # type: (str, str, List[Dict], Optional[Dict]) -> None
        """Add rows under both a key and an ISO 3 country code subkey

        Args:
            key (str): Key to update
            countryiso (str): Country to use as subkey
            rows (List[Dict]): List of dictionaries
            hxltags (Optional[Dict]): HXL tag mapping. Defaults to None.

        Returns:
            None
        """
        return

    def add_additional_json(self, downloader, today=None):
        # type: (Download, Optional[datetime]) -> None
        """Download files and add them under keys defined in the configuration

        Args:
            downloader (Download): Download object for downloading
            today (Optional[datetime]): Value to use for today. Defaults to None (datetime.now()).

        Returns:
            None
        """
        return

    def save(self, **kwargs):
        # type: (Any) -> None
        """Save file

        Args:
            **kwargs: Variables to use when evaluating template arguments

        Returns:
            None
        """
        return



from hdx.location.adminone import AdminOne
from hdx.utilities.dateparse import parse_date
from hdx.utilities.downloader import Download
from hdx.scraper.scrapers import run_scrapers


class TestScraper:
    def test_get_tabular(self, configuration):
        with Download(user_agent='test') as downloader:
            today = parse_date('2020-10-01')
            adminone = AdminOne(configuration)
            population_lookup = dict()
            headers, columns, sources = run_scrapers(configuration, ['AFG'], adminone, 'national', downloader, today=today, scrapers=['population'], population_lookup=population_lookup)
            assert headers == [['Population'], ['#population']]
            assert columns == [{'AFG': 38041754}]
            assert sources == [('#population', '2020-10-01', 'World Bank', 'https://data.humdata.org/organization/world-bank-group')]
            headers, columns, sources = run_scrapers(configuration, ['AFG'], adminone, 'national', downloader, today=today, scrapers=['who'], population_lookup=population_lookup)
            assert headers == [['CasesPer100000', 'DeathsPer100000', 'Cases2Per100000', 'Deaths2Per100000'], ['#affected+infected+per100000', '#affected+killed+per100000', '#affected+infected+2+per100000', '#affected+killed+2+per100000']]
            assert columns == [{'AFG': '96.99'}, {'AFG': '3.41'}, {'AFG': '96.99'}, {'AFG': '3.41'}]
            assert sources == [('#affected+infected+per100000', '2020-08-06', 'WHO', 'tests/fixtures/WHO-COVID-19-global-data.csv'), ('#affected+killed+per100000', '2020-08-06', 'WHO', 'tests/fixtures/WHO-COVID-19-global-data.csv'), ('#affected+infected+2+per100000', '2020-08-06', 'WHO', 'tests/fixtures/WHO-COVID-19-global-data.csv'), ('#affected+killed+2+per100000', '2020-08-06', 'WHO', 'tests/fixtures/WHO-COVID-19-global-data.csv')]
            headers, columns, sources = run_scrapers(configuration, ['AFG'], adminone, 'national', downloader, today=today, scrapers=['access'], population_lookup=population_lookup)
            assert headers == [['% of visas pending or denied', '% of travel authorizations or movements denied', 'Number of incidents reported in previous year', 'Number of incidents reported since start of year', 'Number of incidents reported since start of previous year', '% of CERF projects affected by insecurity and inaccessibility', '% of CBPF projects affected by insecurity and inaccessibility', 'Campaign Vaccine', 'Campaign Vaccine Status', 'Number of learners enrolled from pre-primary to tertiary education'], ['#access+visas+pct', '#access+travel+pct', '#event+year+previous+num', '#event+year+todate+num', '#event+year+previous+todate+num', '#activity+cerf+project+insecurity+pct', '#activity+cbpf+project+insecurity+pct', '#service+name', '#status+name', '#population+education']]
            assert columns == [{'AFG': 0.2}, {'AFG': 'N/A'}, {'AFG': '20'}, {'AFG': '2'}, {'AFG': '22'}, {'AFG': 0.5710000000000001}, {'AFG': 0.04}, {'AFG': 'bivalent Oral Poliovirus'}, {'AFG': 'Postponed'}, {'AFG': 9979405}]
            assert sources == [('#access+visas+pct', '2020-10-01', 'OCHA', 'https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv'), ('#access+travel+pct', '2020-10-01', 'OCHA', 'https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv'), ('#event+year+previous+num', '2020-10-01', 'Aid Workers Database', 'https://data.humdata.org/dataset/security-incidents-on-aid-workers'), ('#event+year+todate+num', '2020-10-01', 'Aid Workers Database', 'https://data.humdata.org/dataset/security-incidents-on-aid-workers'), ('#event+year+previous+todate+num', '2020-10-01', 'Aid Workers Database', 'https://data.humdata.org/dataset/security-incidents-on-aid-workers'), ('#activity+cerf+project+insecurity+pct', '2020-10-01', 'UNCERF', 'https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv'), ('#activity+cbpf+project+insecurity+pct', '2020-10-01', 'UNCERF', 'https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv'), ('#service+name', '2020-10-01', 'Multiple sources', 'https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv'), ('#status+name', '2020-10-01', 'Multiple sources', 'https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv'), ('#population+education', '2020-10-01', 'UNESCO', 'https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv')]
            headers, columns, sources = run_scrapers(configuration, ['AFG'], adminone, 'national', downloader, today=today, scrapers=['sadd'], population_lookup=population_lookup)
            assert headers == [['Cases (% male)', 'Cases (% female)', 'Deaths (% male)', 'Deaths (% female)'], ['#affected+infected+m+pct', '#affected+f+infected+pct', '#affected+killed+m+pct', '#affected+f+killed+pct']]
            assert columns == [{'AFG': '0.7044'}, {'AFG': '0.2956'}, {'AFG': '0.7498'}, {'AFG': '0.2502'}]
            assert sources == [('#affected+infected+m+pct', '2020-08-07', 'SADD', 'tests/fixtures/covid-19-sex-disaggregated-data.csv'), ('#affected+f+infected+pct', '2020-08-07', 'SADD', 'tests/fixtures/covid-19-sex-disaggregated-data.csv'), ('#affected+killed+m+pct', '2020-08-07', 'SADD', 'tests/fixtures/covid-19-sex-disaggregated-data.csv'), ('#affected+f+killed+pct', '2020-08-07', 'SADD', 'tests/fixtures/covid-19-sex-disaggregated-data.csv')]
            headers, columns, sources = run_scrapers(configuration, ['AFG'], adminone, 'subnational', downloader, today=today, scrapers=['gam'], population_lookup=population_lookup)
            assert headers == [['Malnutrition Estimate'], ['#severity+malnutrition+num+subnational']]
            assert columns == [{'AF17': 3.371688, 'AF31': 3.519166, 'AF09': 1.524646, 'AF21': 1.319626, 'AF10': 1.40426, 'AF24': 1.043487, 'AF33': 2.745447, 'AF29': 2.478977, 'AF11': 1.022871, 'AF23': 1.340286, 'AF30': 1.677612, 'AF32': 1.687488, 'AF28': 0.6210205, 'AF01': 1.282291, 'AF27': 1.378641, 'AF02': 3.552082, 'AF14': 0.7653555, 'AF15': 0.953823, 'AF19': 1.684882, 'AF07': 2.090165, 'AF05': 0.9474334, 'AF06': 2.162038, 'AF34': 1.6455, 'AF16': 1.927783, 'AF12': 4.028857, 'AF13': 9.150105, 'AF08': 1.64338, 'AF03': 2.742952, 'AF20': 1.382376, 'AF22': 1.523334, 'AF18': 0.9578965, 'AF25': 0.580423, 'AF04': 0.501081, 'AF26': 4.572629}]
            assert sources == [('#severity+malnutrition+num+subnational', '2020-10-01', 'UNICEF', 'tests/fixtures/unicef_who_wb_global_expanded_databases_severe_wasting.xlsx')]



# -*- coding: UTF-8 -*-
"""Global fixtures"""
from collections import UserDict
from os.path import join

import pytest
from hdx.hdx_configuration import Configuration
from hdx.hdx_locations import Locations


@pytest.fixture(scope='session')
def configuration():
    Configuration._create(hdx_read_only=True, hdx_site='prod', user_agent='test',
                          project_config_yaml=join('tests', 'config', 'project_configuration.yml'))
    Locations.set_validlocations(
        [{'name': 'afg', 'title': 'Afghanistan'}, {'name': 'pse', 'title': 'State of Palestine'}])
    return Configuration.read()


@pytest.fixture(scope='session')
def fixtures():
    return join('tests', 'fixtures')



# -*- coding: utf-8 -*-
import logging
from datetime import datetime
from os.path import join
from typing import Any, Dict, Tuple, List, Iterator, Union, Optional

from hdx.data.dataset import Dataset
from hdx.utilities.downloader import Download
from hdx.utilities.path import temp_dir
from jsonpath_ng import parse
from olefile import olefile

from hdx.scraper import get_date_from_dataset_date, match_template

logger = logging.getLogger(__name__)


def get_url(url, **kwargs):
    # type: (str, Any) -> str
    """Get url from a string replacing any template arguments

    Args:
        url (str): Url to read
        **kwargs: Variables to use when evaluating template arguments

    Returns:
        str: Url with any template arguments replaced
    """
    for kwarg in kwargs:
        exec('%s="%s"' % (kwarg, kwargs[kwarg]))
    template_string, match_string = match_template(url)
    if template_string:
        replace_string = eval(match_string)
        url = url.replace(template_string, replace_string)
    return url


def read_tabular(downloader, datasetinfo, **kwargs):
    # type: (Download, Dict, Any) -> Tuple[List[str],Iterator[Union[List,Dict]]]
    """Read data from tabular source eg. csv, xls, xlsx

    Args:
        downloader (Download): Download object for downloading files
        datasetinfo (Dict): Dictionary of information about dataset
        **kwargs: Variables to use when evaluating template arguments

    Returns:
        Tuple[List[str],Iterator[Union[List,Dict]]]: Tuple (headers, iterator where each row is a list or dictionary)
    """
    url = get_url(datasetinfo['url'], **kwargs)
    sheet = datasetinfo.get('sheet')
    headers = datasetinfo.get('headers')
    if headers is None:
        headers = 1
        datasetinfo['headers'] = 1
    if isinstance(headers, list):
        kwargs['fill_merged_cells'] = True
    format = datasetinfo['format']
    compression = datasetinfo.get('compression')
    if compression:
        kwargs['compression'] = compression
    return downloader.get_tabular_rows(url, sheet=sheet, headers=headers, dict_form=True, format=format, **kwargs)


def read_ole(downloader, datasetinfo, **kwargs):
    # type: (Download, Dict, Any) -> Tuple[List[str],Iterator[Union[List,Dict]]]
    """Read data from OLE Excel source

    Args:
        downloader (Download): Download object for downloading files
        datasetinfo (Dict): Dictionary of information about dataset
        **kwargs: Variables to use when evaluating template arguments

    Returns:
        Tuple[List[str],Iterator[Union[List,Dict]]]: Tuple (headers, iterator where each row is a list or dictionary)
    """
    url = get_url(datasetinfo['url'], **kwargs)
    with temp_dir('ole') as folder:
        path = downloader.download_file(url, folder, 'olefile')
        ole = olefile.OleFileIO(path)
        data = ole.openstream('Workbook').getvalue()
        outputfile = join(folder, 'excel_file.xls')
        with open(outputfile, 'wb') as f:
            f.write(data)
        datasetinfo['url'] = outputfile
        datasetinfo['format'] = 'xls'
        return read_tabular(downloader, datasetinfo, **kwargs)


def read_json(downloader, datasetinfo, **kwargs):
    # type: (Download, Dict, Any) -> Optional[Iterator[Union[List,Dict]]]
    """Read data from json source allowing for JSONPath expressions

    Args:
        downloader (Download): Download object for downloading JSON
        datasetinfo (Dict): Dictionary of information about dataset
        **kwargs: Variables to use when evaluating template arguments

    Returns:
        Optional[Iterator[Union[List,Dict]]]: Iterator or None
    """
    url = get_url(datasetinfo['url'], **kwargs)
    response = downloader.download(url)
    json = response.json()
    expression = datasetinfo.get('jsonpath')
    if expression:
        expression = parse(expression)
        json = expression.find(json)
    if isinstance(json, list):
        return iter(json)
    return None


def read_hdx_metadata(datasetinfo, today=None):
    # type: (Dict, Optional[datetime]) -> None
    """Read metadata from HDX dataset and add to input dictionary

    Args:
        datasetinfo (Dict): Dictionary of information about dataset
        today (Optional[datetime]): Value to use for today. Defaults to None (datetime.now()).

    Returns:
        None
    """
    dataset_name = datasetinfo['dataset']
    dataset = Dataset.read_from_hdx(dataset_name)
    format = datasetinfo['format']
    url = datasetinfo.get('url')
    if not url:
        for resource in dataset.get_resources():
            if resource['format'] == format.upper():
                url = resource['url']
                break
        if not url:
            raise ValueError('Cannot find %s resource in %s!' % (format, dataset_name))
        datasetinfo['url'] = url
    if 'date' not in datasetinfo:
        datasetinfo['date'] = get_date_from_dataset_date(dataset, today=today)
    if 'source' not in datasetinfo:
        datasetinfo['source'] = dataset['dataset_source']
    if 'source_url' not in datasetinfo:
        datasetinfo['source_url'] = dataset.get_hdx_url()


def read_hdx(downloader, datasetinfo, today=None):
    # type: (Download, Dict, Optional[datetime]) -> Tuple[List[str],Iterator[Union[List,Dict]]]
    """Read data and metadata from HDX dataset

    Args:
        downloader (Download): Download object for downloading files
        datasetinfo (Dict): Dictionary of information about dataset
        **kwargs: Variables to use when evaluating template arguments

    Returns:
        Tuple[List[str],Iterator[Union[List,Dict]]]: Tuple (headers, iterator where each row is a list or dictionary)
    """
    read_hdx_metadata(datasetinfo, today=today)
    return read_tabular(downloader, datasetinfo)


def read(downloader, datasetinfo, today=None, **kwargs):
    # type: (Download, Dict, Optional[datetime], Any) -> Tuple[List[str],Iterator[Union[List,Dict]]]
    """Read data and metadata from HDX dataset

    Args:
        downloader (Download): Download object for downloading files
        datasetinfo (Dict): Dictionary of information about dataset
        today (Optional[datetime]): Value to use for today. Defaults to None (datetime.now()).
        **kwargs: Variables to use when evaluating template arguments in urls

    Returns:
        Tuple[List[str],Iterator[Union[List,Dict]]]: Tuple (headers, iterator where each row is a list or dictionary)
    """
    format = datasetinfo['format']
    if format == 'json':
        if 'dataset' in datasetinfo:
            read_hdx_metadata(datasetinfo, today=today)
        iterator = read_json(downloader, datasetinfo, **kwargs)
        headers = None
    elif format == 'ole':
        headers, iterator = read_ole(downloader, datasetinfo, **kwargs)
    elif format in ['csv', 'xls', 'xlsx']:
        if 'dataset' in datasetinfo:
            headers, iterator = read_hdx(downloader, datasetinfo, today=today)
        else:
            headers, iterator = read_tabular(downloader, datasetinfo, **kwargs)
    else:
        raise ValueError('Invalid format %s for %s!' % (format, datasetinfo['name']))
    return headers, iterator



# -*- coding: utf-8 -*-
import re
from datetime import datetime
from typing import Tuple, Optional, Any, Dict, Union, List

from hdx.data.dataset import Dataset

template = re.compile('{{.*?}}')


def match_template(input):
    # type: (str) -> Tuple[Optional[str], Optional[str]]
    """Try to match {{XXX}} in input string

    Args:
        input (str): String in which to look for template

    Returns:
        Tuple[Optional[str], Optional[str]]: (Matched string with brackets, matched string without brackets)
    """
    match = template.search(input)
    if match:
        template_string = match.group()
        return template_string, template_string[2:-2]
    return None, None


def get_rowval(row, valcol):
    # type: (Dict, str) -> Any
    """Get the value of a particular column in a row expanding any template it contains

    Args:
        row (Dict): Dictionary
        valcol (str): Column which may be a string in which to look for template

    Returns:
        Any: Value of column or template
    """
    if '{{' in valcol:
        repvalcol = valcol
        for match in template.finditer(valcol):
            template_string = match.group()
            replace_string = 'row["%s"]' % template_string[2:-2]
            repvalcol = repvalcol.replace(template_string, replace_string)
        return eval(repvalcol)
    else:
        result = row[valcol]
        if isinstance(result, str):
            return result.strip()
        return result


def get_date_from_dataset_date(dataset, today=None):
    # type: (Union[Dataset, str], datetime) -> Optional[str]
    """Return the date or end date of a dataset

    Args:
        dataset (Union[Dataset, str]): Dataset object or name or id of dataset
        today (datetime): Date to use for today. Defaults to None (datetime.now())

    Returns:
        Any: Value of column or template
    """
    if isinstance(dataset, str):
        dataset = Dataset.read_from_hdx(dataset)
    if today is None:
        date_info = dataset.get_date_of_dataset()
    else:
        date_info = dataset.get_date_of_dataset(today=today)
    enddate = date_info.get('enddate_str')
    if not enddate:
        return None
    return enddate[:10]


def add_population(population_lookup, headers, columns):
    # type: (Dict, List[str], List[Dict]) -> None
    """Add population data to dictionary

    Args:
        population_lookup (Dict): Population dictionary
        headers (List[str]): List of headers
        columns (List[Dict]): List of columns

    Returns:
        None
    """
    if population_lookup is None:
        return
    try:
        population_index = headers[1].index('#population')
    except ValueError:
        population_index = None
    if population_index is not None:
        for key, value in columns[population_index].items():
            try:
                valint = int(value)
                population_lookup[key] = valint
            except ValueError:
                pass
